﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Media.Imaging;

namespace WpfGeoTagger.Models
{
	/// <summary>
	/// 住所を表します。
	/// </summary>
	class Address
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		private Address()
		{
		}

		/// <summary>
		/// 指定した Object が、現在の Object と等しいかどうかを判断します。
		/// </summary>
		/// <param name="obj">現在の Object と比較する Object。</param>
		/// <returns>指定した Object が現在の Object と等しい場合は true。それ以外の場合は false。</returns>
		public override bool Equals( object obj )
		{
			if( obj == null || this.GetType() != obj.GetType() ) { return false; }

			var address = obj as Address;
			if( String.Compare( this.FormatedAddress, address.FormatedAddress, true ) != 0 ||
				String.Compare( this.Country,         address.Country,         true ) != 0 ||
				String.Compare( this.ProvinceState,   address.ProvinceState,   true ) != 0 ||
				String.Compare( this.City,            address.City,            true ) != 0 )
			{
				return false;
			}

			return true;
		}

		/// <summary>
		/// 特定の型のハッシュ関数として機能します。
		/// </summary>
		/// <returns>現在の Object のハッシュ コード。</returns>
		public override int GetHashCode()
		{
			return ( this.FormatedAddress.GetHashCode() ^ this.Country.GetHashCode() ^ this.Country.GetHashCode() ^ this.City.GetHashCode() );
		}

		/// <summary>
		/// 現在の Address を表す System.String を返します。
		/// </summary>
		/// <returns>GeoTag の内容を表す System.String。</returns>
		public override string ToString()
		{
			return String.Format( "住所 = {0}", this.FormatedAddress );
		}
		
		/// <summary>
		/// 現在の Address をメタデータへ書き込みます。
		/// </summary>
		/// <param name="metadata">画像のメタデータを表すオブジェクト。</param>
		public void Write( BitmapMetadata metadata )
		{
			//this.WriteIptc( metadata );
			this.WriteXmp( metadata );
		}

		/// <summary>
		/// 現在の Address を IPTC としてメタデータへ書き込みます。
		/// </summary>
		/// <param name="metadata">画像のメタデータを表すオブジェクト。</param>
		private void WriteIptc( BitmapMetadata metadata )
		{
			metadata.SetQuery( Address.IptcPathContentLocationName, this.FormatedAddress );

			// 国
			if( String.IsNullOrEmpty( this.Country ) )
			{
				metadata.RemoveQuery( Address.IptcPathCountry );
			}
			else
			{
				metadata.SetQuery( Address.IptcPathCountry, this.Country );
			}

			// 都道府県や州
			if( String.IsNullOrEmpty( this.ProvinceState ) )
			{
				metadata.RemoveQuery( Address.IptcPathProvinceState );
			}
			else
			{
				metadata.SetQuery( Address.IptcPathProvinceState, this.ProvinceState );
			}

			// 都市
			if( String.IsNullOrEmpty( this.City ) )
			{
				metadata.RemoveQuery( Address.IptcPathCity );
			}
			else
			{
				metadata.SetQuery( Address.IptcPathCity, this.City );
			}
		}

		/// <summary>
		/// 現在の Address を XMP としてメタデータへ書き込みます。
		/// </summary>
		/// <param name="metadata">画像のメタデータを表すオブジェクト。</param>
		private void WriteXmp( BitmapMetadata metadata )
		{
			metadata.SetQuery( Address.XmpPathContentLocationName, this.FormatedAddress );

			// 国
			if( String.IsNullOrEmpty( this.Country ) )
			{
				metadata.RemoveQuery( Address.XmpPathCountry );
			}
			else
			{
				metadata.SetQuery( Address.XmpPathCountry, this.Country );
			}

			// 都道府県や州
			if( String.IsNullOrEmpty( this.ProvinceState ) )
			{
				metadata.RemoveQuery( Address.XmpPathProvinceState );
			}
			else
			{
				metadata.SetQuery( Address.XmpPathProvinceState, this.ProvinceState );
			}

			// 都市
			if( String.IsNullOrEmpty( this.City ) )
			{
				metadata.RemoveQuery( Address.XmpPathCity );
			}
			else
			{
				metadata.SetQuery( Address.XmpPathCity, this.City );
			}
		}

		/// <summary>
		/// JavaScript によって、Goole Map の Geocoder から取得した住所情報から Address インスタンスを生成します。
		/// </summary>
		/// <param name="address">Geocoder から取得した住所文字列。</param>
		/// <returns></returns>
		public static Address FromGeocoderAddress( string address )
		{
			if( String.IsNullOrEmpty( address ) ) { return null; }

			// 整形済みの住所は常に存在している必要がある
			var formatedAddress = Address.NormlizeMultiByteNumberSymbol( Address.GetAdressValue( address, Address.TagFormatedAddress ) );
			if( String.IsNullOrEmpty( formatedAddress ) ) { return null; }

			return new Address()
			{
				FormatedAddress = formatedAddress,
				Country = Address.GetAdressValue( address, Address.TagCountry ),
				ProvinceState = Address.GetAdressValue( address, Address.TagProvinceState ),
				City = Address.GetAddressValueCity( address )
			};
		}

		/// <summary>
		/// 画像のメタデータから Address インスタンスを生成します。
		/// </summary>
		/// <param name="metadata">画像のメタデータを表すオブジェクト。</param>
		/// <returns>成功時は Address インスタンス。失敗時は null 参照。</returns>
		public static Address FromMetadata( BitmapMetadata metadata )
		{
			Address address = null;
			{
				var formatedAddress = metadata.GetQuery( Address.IptcPathContentLocationName ) as string;
				var country         = metadata.GetQuery( Address.IptcPathCountry ) as string;
				var state           = metadata.GetQuery( Address.IptcPathProvinceState ) as string;
				var city            = metadata.GetQuery( Address.IptcPathCity ) as string;

				if( !String.IsNullOrEmpty( formatedAddress ) ||
					!String.IsNullOrEmpty( country ) ||
					!String.IsNullOrEmpty( state ) ||
					!String.IsNullOrEmpty( city ) )
				{
					address = new Address() { FormatedAddress = formatedAddress, Country = country, ProvinceState = state, City = city };
				}
			}

			return address;
		}

		/// <summary>
		/// 文字列中の全角の数値や記号を半角に変換する。
		/// </summary>
		/// <param name="str">文字列。</param>
		/// <returns>変換後の文字列。</returns>
		private static string NormlizeMultiByteNumberSymbol( string str )
		{
			if( String.IsNullOrEmpty( str ) ) { return str; }

			StringBuilder sb = new StringBuilder();
			for( int i = 0; i < str.Length; ++i )
			{
				switch( str[ i ] )
				{
				case '０': sb.Append( '0' ); break;
				case '１': sb.Append( '1' ); break;
				case '２': sb.Append( '2' ); break;
				case '３': sb.Append( '3' ); break;
				case '４': sb.Append( '4' ); break;
				case '５': sb.Append( '5' ); break;
				case '６': sb.Append( '6' ); break;
				case '７': sb.Append( '7' ); break;
				case '８': sb.Append( '8' ); break;
				case '９': sb.Append( '9' ); break;
				case '−': sb.Append( '-' ); break;	// ハイフン ( マイナスと同じに見えるが、文字コードが違う )
				case '－': sb.Append( '-' ); break;
				case '＋': sb.Append( '+' ); break;
				case '／': sb.Append( '/' ); break;
				case '＊': sb.Append( '*' ); break;
				default:
					sb.Append( str[ i ] );
					break;
				}
			}

			return sb.ToString();
		}

		/// <summary>
		/// JavaScript によって、Goole Map の Geocoder から取得した住所情報から、一つの住所区分を取得します。
		/// 住所情報となる文字列は、"country:日本;administrative_area_level_1:東京都; ...etc" という形式となります。
		/// </summary>
		/// <param name="address">Geocoder から取得した住所情報の文字列。</param>
		/// <param name="tag">取得する値の種別を示すタグ名。</param>
		/// <returns>取得した住所区分。</returns>
		private static string GetAdressValue( string address, string tag )
		{
			return Address.GetAdressValue( address, tag, true );
		}

		/// <summary>
		/// JavaScript によって、Goole Map の Geocoder から取得した住所情報から、一つの住所区分を取得します。
		/// 住所情報となる文字列は、"country:日本;administrative_area_level_1:東京都; ...etc" という形式となります。
		/// </summary>
		/// <param name="address">Geocoder から取得した住所情報の文字列。</param>
		/// <param name="tag">取得する値の種別を示すタグ名。</param>
		/// <param name="isLastIndexOf">tag を先頭から検索する場合は true、末尾からの場合は false。</param>
		/// <returns>取得した住所区分。</returns>
		private static string GetAdressValue( string address, string tag, bool isIndexOf )
		{
			int index = ( isIndexOf ? address.IndexOf( tag ) : address.LastIndexOf( tag ) );
			if( index == -1 ) { return String.Empty; }

			int begin = index + tag.Length;
			index = address.IndexOf( ";", begin );
			if( index == -1 ) { return String.Empty; }

			int length = index - begin;
			return address.Substring( begin, length );
		}

		/// <summary>
		/// 都市情報を取得します。
		/// </summary>
		/// <param name="address">Geocoder から取得した住所情報の文字列。</param>
		/// <returns></returns>
		private static string GetAddressValueCity( string address )
		{
			// 例えば住所の都市部分が A 市 B 町 6-2 の場合、{ "2", "6", "B 町", "A 市" } のように格納されていて、
			// "A 市" が TagCity、それ以降が全て TagSubCity になるので、TagCity と最後に登場する TagSubCity で取得する。
			//
			var city = String.Format( "{0}{1}", Address.GetAdressValue( address, Address.TagCity ), Address.GetAdressValue( address, Address.TagSubCity, false ) );
			if( String.IsNullOrEmpty( city ) )
			{
				city = String.Format( "{0}{1}", Address.GetAdressValue( address, Address.TagProvinceState2 ), Address.GetAdressValue( address, Address.TagProvinceState3 ) );
			}

			return city;
		}

		#region プロパティ

		/// <summary>
		/// 都市を取得します。
		/// </summary>
		public string City { get; private set; }

		/// <summary>
		/// 国を取得します。
		/// </summary>
		public string Country { get; private set; }

		/// <summary>
		/// Geocoder によって整形済みの住所を取得します。
		/// </summary>
		public string FormatedAddress { get; private set; }

		/// <summary>
		/// 県や州を取得します。
		/// </summary>
		public string ProvinceState { get; private set; }

		#endregion

		#region 定数

		/// <summary>
		/// JavaScript で埋め込んだ住所情報の「整形済み住所」を表すタグ文字列。
		/// </summary>
		private const string TagFormatedAddress = "[formatted_address]";

		/// <summary>
		/// JavaScript で埋め込んだ住所情報の「国」を表すタグ文字列。
		/// </summary>
		private const string TagCountry = "[country]";

		/// <summary>
		/// JavaScript で埋め込んだ住所情報の「都道府県や州」を表すタグ文字列。
		/// </summary>
		private const string TagProvinceState = "[administrative_area_level_1]";

		/// <summary>
		/// JavaScript で埋め込んだ住所情報の「都市」を表すタグ文字列。
		/// TagCity - TagSubCity で都市情報を取得出来なかった場合に使用する。
		/// </summary>
		private const string TagProvinceState2 = "[administrative_area_level_2]";

		/// <summary>
		/// JavaScript で埋め込んだ住所情報の「町」を表すタグ文字列。
		/// TagCity - TagSubCity で都市情報を取得出来なかった場合に使用する。
		/// </summary>
		private const string TagProvinceState3 = "[administrative_area_level_3]";

		/// <summary>
		/// JavaScript で埋め込んだ住所情報の「都市」を表すタグ文字列。
		/// </summary>
		private const string TagCity = "[locality]";

		/// <summary>
		/// JavaScript で埋め込んだ住所情報の「町」を表すタグ文字列。
		/// </summary>
		private const string TagSubCity = "[sublocality]";

		/// <summary>
		/// JavaScript で埋め込んだ住所情報の、一つの種別の終端を示す文字列。
		/// </summary>
		private const string TagValueEnd = ";";

		/// <summary>
		/// IPTC 情報における「撮影場所」を示すパス文字列。
		/// </summary>
		private const string IptcPathContentLocationName = "/app13/irb/8bimiptc/iptc/content location name";

		/// <summary>
		/// IPTC 情報における「撮影された国」を示すパス文字列。
		/// </summary>
		private const string IptcPathCountry = "/app13/irb/8bimiptc/iptc/country\\/primary location name";

		/// <summary>
		/// IPTC 情報における「撮影された都道府県や州」を示すパス文字列。
		/// </summary>
		private const string IptcPathProvinceState = "/app13/irb/8bimiptc/iptc/province\\/state";

		/// <summary>
		/// IPTC 情報における「撮影された都市」を示すパス文字列。
		/// </summary>
		private const string IptcPathCity = "/app13/irb/8bimiptc/iptc/city";

		/// <summary>
		/// XMP 情報における「撮影場所」を示すパス文字列。
		/// </summary>
		private const string XmpPathContentLocationName = "/xmp/Iptc4xmpCore:Location";

		/// <summary>
		/// XMP 情報における「撮影された国」を示すパス文字列。
		/// </summary>
		private const string XmpPathCountry = "/xmp/photoshop:Country";

		/// <summary>
		/// XMP 情報における「撮影された都道府県や州」を示すパス文字列。
		/// </summary>
		private const string XmpPathProvinceState = "/xmp/photoshop:State";

		/// <summary>
		/// XMP 情報における「撮影された都市」を示すパス文字列。
		/// </summary>
		private const string XmpPathCity = "/xmp/photoshop:City";
	
		#endregion
	}
}
