﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.Windows.Forms;
using System.Windows.Input;
using WpfGeoTagger.Commands;

namespace WpfGeoTagger.ViewModels
{
	/// <summary>
	/// MainWindow の Model と View を仲介します。
	/// </summary>
	class MainWindowViewModel : ViewModelBase
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="uri">ブラウザに表示するページの URI。</param>
		/// <param name="invokScript">ブラウザのスクリプト実行を行う為のデリゲート。</param>
		public MainWindowViewModel( string uri )
		{
			this.MapViewModel    = new MapViewModel( uri );
			this.Images          = new ObservableCollection< ImageViewModel >();
			this.AddImageCommand = new DelegateCommand( OnAddImageCommand );

			this.MapViewModel.PropertyChanged += new PropertyChangedEventHandler( OnMapViewModelPropertyChanged );
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void OnMapViewModelPropertyChanged( object sender, PropertyChangedEventArgs e )
		{
			if( e.PropertyName == "SelectedImage" )
			{
				this.NotifyPropertyChanged( e.PropertyName );
			}
		}

		/// <summary>
		/// AddImageCommand が実行された時に呼び出されます。
		/// </summary>
		private void OnAddImageCommand()
		{
			using( var dialog = new OpenFileDialog() { Filter = "JPEG (*.jpg;*jpeg)|*.jpg;*jpeg", Multiselect = true } )
			{
				if( DialogResult.OK != dialog.ShowDialog() ) { return; }

				foreach( var fileName in dialog.FileNames )
				{
					if( this._registeredImages.Contains( fileName ) ) { continue; }

					try
					{
						this.Images.Add( new ImageViewModel( MainWindowViewModel.NextId, fileName, OnMetadataLoadedd ) );
						
						// 画像の読み込みに成功したら、重複防止と識別子の調整を行う
						this._registeredImages.Add( fileName );
						++MainWindowViewModel.NextId;
					}
					catch( Exception exp )
					{
						Debug.WriteLine( exp.Message );
					}
				}
			}
		}
		
		/// <summary>
		/// メタデータの読み込みが完了した時に呼び出されます。
		/// </summary>
		/// <param name="image">メタデータを読み込んだ画像。</param>
		private void OnMetadataLoadedd( ImageViewModel image )
		{
			if( image.Metadata.GeoTag != null )
			{
				this.MapViewModel.AddMarker( image );
			}
		}

		/// <summary>
		/// 選択されている画像を取得または設定します。
		/// </summary>
		public ImageViewModel SelectedImage
		{
			get
			{
				return this.MapViewModel.SelectedImage;
			}
			set
			{
				this.MapViewModel.SelectedImage = value;
			}
		}

		/// <summary>
		/// 画像の追加コマンドを取得します。
		/// </summary>
		public ICommand AddImageCommand { get; private set; }

		/// <summary>
		/// 画像の Model と View を仲介するオブジェクトのコレクションを取得します。
		/// </summary>
		public ObservableCollection< ImageViewModel > Images { get; private set; }

		/// <summary>
		/// Map の Model と View を仲介するオブジェクトを取得します。
		/// </summary>
		public MapViewModel MapViewModel { get; private set; }

		/// <summary>
		/// 登録済みの画像ファイルのパスを調べる為のコレクション。
		/// </summary>
		private HashSet< string > _registeredImages = new HashSet< string >( StringComparer.OrdinalIgnoreCase );
	
		/// <summary>
		/// 次に割り当てる画像の識別子。
		/// </summary>
		private static int NextId = 0;
	}
}
